package explorviz.live_trace_processing.filter.reconstruction;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentSkipListMap;

import explorviz.live_trace_processing.Constants;
import explorviz.live_trace_processing.filter.AbstractFilter;
import explorviz.live_trace_processing.filter.reduction.ITraceReduction;
import explorviz.live_trace_processing.reader.TimeProvider;
import explorviz.live_trace_processing.record.IRecord;
import explorviz.live_trace_processing.record.event.AbstractEventRecord;
import explorviz.live_trace_processing.record.misc.TerminateRecord;
import explorviz.live_trace_processing.record.misc.TimedPeriodRecord;
import explorviz.live_trace_processing.record.trace.Trace;

public final class TraceReconstructionFilter extends AbstractFilter implements ITraceReconstruction {
	private final long maxTraceTimeout;

	private final Map<Long, TraceReconstructionBuffer> traceId2trace = new ConcurrentSkipListMap<Long, TraceReconstructionBuffer>();

	public TraceReconstructionFilter(final long maxTraceTimeout,
			final ITraceReduction traceReduction) {
		super(traceReduction, Constants.TRACE_RECONSTRUCTION_DISRUPTOR_SIZE,
				Constants.TRACE_RECONSTRUCTION_OUTPUT_BUFFER_SIZE, "Reconstructed traces/sec");
		this.maxTraceTimeout = maxTraceTimeout;
	}

	@Override
	public final void processRecord(final IRecord record) {
		if (record instanceof AbstractEventRecord) {
			final AbstractEventRecord abstractOperationEvent = ((AbstractEventRecord) record);

			final long traceId = abstractOperationEvent.getTraceId();
			final TraceReconstructionBuffer traceBuffer = getBufferForTraceId(abstractOperationEvent
					.getTraceId());
			traceBuffer.insertEvent(abstractOperationEvent);

			if (traceBuffer.isFinished()) {
				deliver(traceBuffer.toTrace(true));
				traceId2trace.remove(traceId);
			}
		} else if (record instanceof Trace) {
			final Trace trace = (Trace) record;
			if (trace.isValid()) {
				deliver(trace);
			} else {
				getBufferForTraceId(trace.getTraceEvents().get(0).getTraceId());
			}
		} else if (record instanceof TimedPeriodRecord) {
			checkForTimeouts(TimeProvider.getCurrentTimestamp());
			periodicFlush(record);
			// deliver(record);
		} else if (record instanceof TerminateRecord) {
			terminate();
			deliver(record);
		} else {
			deliver(record);
		}
	}

	private final TraceReconstructionBuffer getBufferForTraceId(final long traceId) {
		TraceReconstructionBuffer traceBuffer = traceId2trace.get(traceId);
		if (traceBuffer == null) {
			traceBuffer = new TraceReconstructionBuffer();
			traceId2trace.put(traceId, traceBuffer);
		}
		return traceBuffer;
	}

	private void checkForTimeouts(final long timestamp) {
		final long traceTimeout = timestamp - maxTraceTimeout;
		final List<Long> traceIdsToRemove = new ArrayList<Long>();

		for (final Entry<Long, TraceReconstructionBuffer> entry : traceId2trace.entrySet()) {
			final TraceReconstructionBuffer traceBuffer = entry.getValue();
			if (traceBuffer.isUpdatedInThisPeriod()) {
				traceBuffer.setUpdatedInThisPeriod(false);
			} else {
				traceBuffer.updateLastBufferInsert();

				if ((timestamp - traceBuffer.getLastBufferInsert()) <= traceTimeout) {
					deliver(traceBuffer.toTrace(false));
					traceIdsToRemove.add(entry.getKey());
				}
			}
		}

		for (final long traceIdToRemove : traceIdsToRemove) {
			traceId2trace.remove(traceIdToRemove);
		}
	}

	private void terminate() {
		for (final TraceReconstructionBuffer entry : traceId2trace.values()) {
			deliver(entry.toTrace(false));
		}
		traceId2trace.clear();
	}
}