package explorviz.live_trace_processing.filter.reduction.summarization;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentSkipListMap;

import explorviz.live_trace_processing.Constants;
import explorviz.live_trace_processing.filter.AbstractFilter;
import explorviz.live_trace_processing.filter.IPipeReceiver;
import explorviz.live_trace_processing.filter.reduction.ITraceReduction;
import explorviz.live_trace_processing.reader.TimeProvider;
import explorviz.live_trace_processing.record.IRecord;
import explorviz.live_trace_processing.record.misc.TerminateRecord;
import explorviz.live_trace_processing.record.misc.TimedPeriodRecord;
import explorviz.live_trace_processing.record.trace.Trace;
import explorviz.live_trace_processing.record.trace.TraceComperator;

public class TracePatternSummarizationFilter extends AbstractFilter implements ITraceReduction {
	private final long maxCollectionDuration;

	private final Map<Trace, TracePatternSummarizationBuffer> trace2buffer = new ConcurrentSkipListMap<Trace, TracePatternSummarizationBuffer>(
			new TraceComperator());

	// TODO cache for buffer objects

	public TracePatternSummarizationFilter(final long maxCollectionDuration,
			final IPipeReceiver sinkReceiver) {
		super(sinkReceiver, Constants.TRACE_SUMMARIZATION_DISRUPTOR_SIZE,
				Constants.TRACE_SUMMARIZATION_OUTPUT_BUFFER_SIZE, "Reduced traces/sec");

		this.maxCollectionDuration = maxCollectionDuration;
	}

	@Override
	public void processRecord(final IRecord record) {
		if (record instanceof Trace) {
			final Trace trace = (Trace) record;
			if (trace.isValid()) {
				insertIntoBuffer(trace);
			} else {
				deliver(trace);
			}
		} else if (record instanceof TimedPeriodRecord) {
			processTimeoutQueue(TimeProvider.getCurrentTimestamp());
			periodicFlush(record);
			// deliver(record);
		} else if (record instanceof TerminateRecord) {
			terminate();
			deliver(record);
		} else {
			deliver(record);
		}
	}

	private void insertIntoBuffer(final Trace trace) {
		TracePatternSummarizationBuffer traceAggregationBuffer = trace2buffer.get(trace);
		if (traceAggregationBuffer == null) {
			traceAggregationBuffer = new TracePatternSummarizationBuffer(
					TimeProvider.getCurrentTimestamp());
			trace2buffer.put(trace, traceAggregationBuffer);
		}
		traceAggregationBuffer.insertTrace(trace);
	}

	private void processTimeoutQueue(final long timestamp) {
		final long bufferTimeout = timestamp - maxCollectionDuration;
		final List<Trace> toRemove = new ArrayList<Trace>();
		for (final TracePatternSummarizationBuffer traceBuffer : trace2buffer.values()) {
			if (traceBuffer.getBufferCreatedTimestamp() <= bufferTimeout) {
				final Trace aggregatedTrace = traceBuffer.getAggregatedTrace();
				deliver(aggregatedTrace);
				toRemove.add(aggregatedTrace);
			}
		}
		for (final Trace traceEventRecords : toRemove) {
			trace2buffer.remove(traceEventRecords);
		}
	}

	private void terminate() {
		for (final TracePatternSummarizationBuffer traceBuffer : trace2buffer.values()) {
			deliver(traceBuffer.getAggregatedTrace());
		}
		trace2buffer.clear();
	}
}
