package explorviz.live_trace_processing.reader;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.SocketChannel;
import java.util.ArrayList;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.lmax.disruptor.RingBuffer;

import explorviz.live_trace_processing.Constants;
import explorviz.live_trace_processing.IdNotAvailableException;
import explorviz.live_trace_processing.StringRegistry;
import explorviz.live_trace_processing.filter.RecordArrayEvent;
import explorviz.live_trace_processing.record.IRecord;
import explorviz.live_trace_processing.record.event.AbstractEventRecord;
import explorviz.live_trace_processing.record.event.constructor.AfterConstructorEventRecord;
import explorviz.live_trace_processing.record.event.constructor.AfterFailedConstructorEventRecord;
import explorviz.live_trace_processing.record.event.constructor.BeforeConstructorEventRecord;
import explorviz.live_trace_processing.record.event.normal.AfterFailedOperationEventRecord;
import explorviz.live_trace_processing.record.event.normal.AfterOperationEventRecord;
import explorviz.live_trace_processing.record.event.normal.BeforeOperationEventRecord;
import explorviz.live_trace_processing.record.event.remote.ReceivedRemoteCallRecord;
import explorviz.live_trace_processing.record.event.remote.SentRemoteCallRecord;
import explorviz.live_trace_processing.record.event.remote.UnknownReceivedRemoteCallRecord;
import explorviz.live_trace_processing.record.event.statics.AfterFailedStaticOperationEventRecord;
import explorviz.live_trace_processing.record.event.statics.AfterStaticOperationEventRecord;
import explorviz.live_trace_processing.record.event.statics.BeforeStaticOperationEventRecord;
import explorviz.live_trace_processing.record.misc.StringRegistryRecord;
import explorviz.live_trace_processing.record.misc.SystemMonitoringRecord;
import explorviz.live_trace_processing.record.trace.HostApplicationMetaDataRecord;
import explorviz.live_trace_processing.record.trace.Trace;

class TCPReaderOneClient extends Thread {

	private static final Logger LOG = LoggerFactory.getLogger(TCPReaderOneClient.class);

	private HostApplicationMetaDataRecord hostApplicationMetadata;

	private final StringRegistry stringRegistry = new StringRegistry(null);
	private final List<byte[]> waitingForStringMessages = new ArrayList<byte[]>(1024);

	private final SocketChannel socketChannel;
	private final RingBuffer<RecordArrayEvent> ringBuffer;

	private IRecord[] outputBuffer = new IRecord[Constants.TCP_READER_OUTPUT_BUFFER_SIZE];
	private int outputBufferIndex = 0;

	public TCPReaderOneClient(final SocketChannel socketChannel,
			final RingBuffer<RecordArrayEvent> ringBuffer) {
		this.socketChannel = socketChannel;
		this.ringBuffer = ringBuffer;
	}

	@Override
	public void run() {
		final ByteBuffer buffer = ByteBuffer.allocateDirect(2 * 1024 * 1024);
		try {
			if (socketChannel.isConnected()) {
				LOG.info("Client " + socketChannel.getRemoteAddress() + " connected.");
			}
			while ((socketChannel.read(buffer)) != -1) {
				buffer.flip();
				messagesfromByteArray(buffer);
			}
		} catch (final IOException ex) {
			LOG.info("Error in read() " + ex.getMessage());
		}
	}

	private final void messagesfromByteArray(final ByteBuffer buffer) {
		while (buffer.remaining() > 0) {
			final byte clazzId = buffer.get();
			switch (clazzId) {
				case HostApplicationMetaDataRecord.CLAZZ_ID: {
					if (buffer.remaining() >= HostApplicationMetaDataRecord.BYTE_LENGTH) {
						readInHostApplicationMetaData(buffer);
						break;
					}
					buffer.position(buffer.position() - 1);
					buffer.compact();
					return;
				}
				case BeforeOperationEventRecord.CLAZZ_ID: {
					if (buffer.remaining() >= BeforeOperationEventRecord.COMPRESSED_BYTE_LENGTH) {
						readInBeforeOperationEvent(buffer);
						break;
					}
					buffer.position(buffer.position() - 1);
					buffer.compact();
					return;
				}
				case AfterFailedOperationEventRecord.CLAZZ_ID: {
					if (buffer.remaining() >= AfterFailedOperationEventRecord.COMPRESSED_BYTE_LENGTH) {
						readInAfterFailedOperationEvent(buffer);
						break;
					}
					buffer.position(buffer.position() - 1);
					buffer.compact();
					return;

				}
				case AfterOperationEventRecord.CLAZZ_ID: {
					if (buffer.remaining() >= AfterOperationEventRecord.COMPRESSED_BYTE_LENGTH) {
						readInAfterOperationEvent(buffer);
						break;
					}
					buffer.position(buffer.position() - 1);
					buffer.compact();
					return;
				}
				case StringRegistryRecord.CLAZZ_ID: {
					int mapId = 0;
					int stringLength = 0;
					if (buffer.remaining() >= 8) {
						mapId = buffer.getInt();
						stringLength = buffer.getInt();
					} else {
						buffer.position(buffer.position() - 1);
						buffer.compact();
						return;
					}

					if (buffer.remaining() >= stringLength) {
						final byte[] stringByteArray = new byte[stringLength];

						buffer.get(stringByteArray);

						stringRegistry.putStringRecord(mapId, new String(stringByteArray));

						checkWaitingMessages();
					} else {
						buffer.position(buffer.position() - 9);
						buffer.compact();
						return;
					}
					break;
				}
				case SystemMonitoringRecord.CLAZZ_ID: {
					if (buffer.remaining() >= SystemMonitoringRecord.BYTE_LENGTH) {
						readInSystemMonitoringRecord(buffer);
						break;
					}
					buffer.position(buffer.position() - 1);
					buffer.compact();
					return;
				}
				case Trace.CLAZZ_ID: {
					if (buffer.remaining() >= Trace.BYTE_LENGTH) {
						final byte validByte = buffer.get();
						boolean valid = true;
						if (validByte == (byte) 0) {
							valid = false;
						}
						final int eventsLength = buffer.getInt();
						final int byteLength = buffer.getInt();
						if ((buffer.remaining() >= byteLength) && (eventsLength > 0)) {
							final List<AbstractEventRecord> events = new ArrayList<AbstractEventRecord>(
									eventsLength);
							for (int i = 0; i < eventsLength; i++) {
								try {
									events.add(AbstractEventRecord.createFromByteBuffer(buffer,
											stringRegistry));
								} catch (final IdNotAvailableException e) {
									// should not happen
									e.printStackTrace();
								}
							}

							putInRingBuffer(new Trace(events, valid));
							break;
						}
						buffer.position(buffer.position() - 10);
						buffer.compact();
						return;
					}
					buffer.position(buffer.position() - 1);
					buffer.compact();
					return;
				}
				case BeforeConstructorEventRecord.CLAZZ_ID: {
					if (buffer.remaining() >= BeforeConstructorEventRecord.COMPRESSED_BYTE_LENGTH) {
						readInBeforeConstructorEvent(buffer);
						break;
					}
					buffer.position(buffer.position() - 1);
					buffer.compact();
					return;
				}
				case AfterFailedConstructorEventRecord.CLAZZ_ID: {
					if (buffer.remaining() >= AfterFailedConstructorEventRecord.COMPRESSED_BYTE_LENGTH) {
						readInAfterFailedConstructorEvent(buffer);
						break;
					}
					buffer.position(buffer.position() - 1);
					buffer.compact();
					return;
				}
				case AfterConstructorEventRecord.CLAZZ_ID: {
					if (buffer.remaining() >= AfterConstructorEventRecord.COMPRESSED_BYTE_LENGTH) {
						readInAfterConstructorEvent(buffer);
						break;
					}
					buffer.position(buffer.position() - 1);
					buffer.compact();
					return;
				}
				case ReceivedRemoteCallRecord.CLAZZ_ID: {
					if (buffer.remaining() >= ReceivedRemoteCallRecord.COMPRESSED_BYTE_LENGTH) {
						readInReceivedRemoteCallEvent(buffer);
						break;
					}
					buffer.position(buffer.position() - 1);
					buffer.compact();
					return;
				}
				case BeforeStaticOperationEventRecord.CLAZZ_ID: {
					if (buffer.remaining() >= BeforeStaticOperationEventRecord.COMPRESSED_BYTE_LENGTH) {
						readInBeforeStaticOperationEvent(buffer);
						break;
					}
					buffer.position(buffer.position() - 1);
					buffer.compact();
					return;
				}
				case AfterFailedStaticOperationEventRecord.CLAZZ_ID: {
					if (buffer.remaining() >= AfterFailedStaticOperationEventRecord.COMPRESSED_BYTE_LENGTH) {
						readInAfterFailedStaticOperationEvent(buffer);
						break;
					}
					buffer.position(buffer.position() - 1);
					buffer.compact();
					return;
				}
				case AfterStaticOperationEventRecord.CLAZZ_ID: {
					if (buffer.remaining() >= AfterStaticOperationEventRecord.COMPRESSED_BYTE_LENGTH) {
						readInAfterStaticOperationEvent(buffer);
						break;
					}
					buffer.position(buffer.position() - 1);
					buffer.compact();
					return;
				}
				case SentRemoteCallRecord.CLAZZ_ID: {
					if (buffer.remaining() >= SentRemoteCallRecord.COMPRESSED_BYTE_LENGTH) {
						readInSentRemoteCallEvent(buffer);
						break;
					}
					buffer.position(buffer.position() - 1);
					buffer.compact();
					return;
				}
				case UnknownReceivedRemoteCallRecord.CLAZZ_ID: {
					if (buffer.remaining() >= UnknownReceivedRemoteCallRecord.COMPRESSED_BYTE_LENGTH) {
						readInUnknownReceivedRemoteCallEvent(buffer);
						break;
					}
					buffer.position(buffer.position() - 1);
					buffer.compact();
					return;
				}
				default: {
					System.out.println("unknown class id " + clazzId + " at offset "
							+ (buffer.position() - 1));
					buffer.clear();
					return;
				}
			}
		}

		buffer.clear();
	}

	private final void readInHostApplicationMetaData(final ByteBuffer buffer) {
		final int hostnameId = buffer.getInt();
		final int applicationId = buffer.getInt();

		try {
			final String hostname = stringRegistry.getStringFromId(hostnameId);
			final String application = stringRegistry.getStringFromId(applicationId);

			hostApplicationMetadata = new HostApplicationMetaDataRecord(hostname, application);
		} catch (final IdNotAvailableException e) {
			putInWaitingMessages(buffer, HostApplicationMetaDataRecord.BYTE_LENGTH_WITH_CLAZZ_ID);
		}
	}

	private final void readInBeforeOperationEvent(final ByteBuffer buffer) {
		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();
		final int objectId = buffer.getInt();
		final int operationId = buffer.getInt();

		try {
			final String operation = stringRegistry.getStringFromId(operationId);

			putInRingBuffer(new BeforeOperationEventRecord(timestamp, traceId, orderIndex,
					objectId, operation, hostApplicationMetadata));
		} catch (final IdNotAvailableException e) {
			putInWaitingMessages(buffer, BeforeOperationEventRecord.COMPRESSED_BYTE_LENGTH + 1);
		}
	}

	private final void readInAfterFailedOperationEvent(final ByteBuffer buffer) {
		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();
		final int causeId = buffer.getInt();

		try {
			final String cause = stringRegistry.getStringFromId(causeId);

			putInRingBuffer(new AfterFailedOperationEventRecord(timestamp, traceId, orderIndex,
					cause, hostApplicationMetadata));
		} catch (final IdNotAvailableException e) {
			putInWaitingMessages(buffer,
					AfterFailedOperationEventRecord.COMPRESSED_BYTE_LENGTH_WITH_CLAZZ_ID);
		}
	}

	private final void readInAfterOperationEvent(final ByteBuffer buffer) {
		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();
		putInRingBuffer(new AfterOperationEventRecord(timestamp, traceId, orderIndex,
				hostApplicationMetadata));
	}

	private final void readInSystemMonitoringRecord(final ByteBuffer buffer) {
		final double cpuUtil = buffer.getDouble();
		final long usedRAM = buffer.getLong();
		final long absoluteRAM = buffer.getLong();

		putInRingBuffer(new SystemMonitoringRecord(cpuUtil, usedRAM, absoluteRAM,
				hostApplicationMetadata));
	}

	private final void readInBeforeConstructorEvent(final ByteBuffer buffer) {
		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();
		final int objectId = buffer.getInt();
		final int operationId = buffer.getInt();

		try {
			final String operation = stringRegistry.getStringFromId(operationId);

			putInRingBuffer(new BeforeConstructorEventRecord(timestamp, traceId, orderIndex,
					objectId, operation, hostApplicationMetadata));
		} catch (final IdNotAvailableException e) {
			putInWaitingMessages(buffer,
					BeforeConstructorEventRecord.COMPRESSED_BYTE_LENGTH_WITH_CLAZZ_ID);
		}
	}

	private final void readInAfterFailedConstructorEvent(final ByteBuffer buffer) {
		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();
		final int causeId = buffer.getInt();

		try {
			putInRingBuffer(new AfterFailedConstructorEventRecord(timestamp, traceId, orderIndex,
					stringRegistry.getStringFromId(causeId), hostApplicationMetadata));
		} catch (final IdNotAvailableException e) {
			putInWaitingMessages(buffer,
					AfterFailedConstructorEventRecord.COMPRESSED_BYTE_LENGTH_WITH_CLAZZ_ID);
		}
	}

	private final void readInAfterConstructorEvent(final ByteBuffer buffer) {
		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();

		putInRingBuffer(new AfterConstructorEventRecord(timestamp, traceId, orderIndex,
				hostApplicationMetadata));
	}

	private final void readInReceivedRemoteCallEvent(final ByteBuffer buffer) {
		final long callerTraceId = buffer.getLong();
		final int callerOrderIndex = buffer.getInt();

		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();

		putInRingBuffer(new ReceivedRemoteCallRecord(timestamp, callerTraceId, callerOrderIndex,
				traceId, orderIndex, hostApplicationMetadata));
	}

	private final void readInBeforeStaticOperationEvent(final ByteBuffer buffer) {
		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();
		final int operationId = buffer.getInt();

		try {
			final String operation = stringRegistry.getStringFromId(operationId);

			putInRingBuffer(new BeforeStaticOperationEventRecord(timestamp, traceId, orderIndex,
					operation, hostApplicationMetadata));
		} catch (final IdNotAvailableException e) {
			putInWaitingMessages(buffer,
					BeforeStaticOperationEventRecord.COMPRESSED_BYTE_LENGTH_WITH_CLAZZ_ID);
		}
	}

	private final void readInAfterFailedStaticOperationEvent(final ByteBuffer buffer) {
		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();
		final int causeId = buffer.getInt();

		try {
			final String cause = stringRegistry.getStringFromId(causeId);

			putInRingBuffer(new AfterFailedStaticOperationEventRecord(timestamp, traceId,
					orderIndex, cause, hostApplicationMetadata));
		} catch (final IdNotAvailableException e) {
			putInWaitingMessages(buffer,
					AfterFailedStaticOperationEventRecord.COMPRESSED_BYTE_LENGTH_WITH_CLAZZ_ID);
		}
	}

	private final void readInAfterStaticOperationEvent(final ByteBuffer buffer) {
		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();

		putInRingBuffer(new AfterStaticOperationEventRecord(timestamp, traceId, orderIndex,
				hostApplicationMetadata));
	}

	private final void readInSentRemoteCallEvent(final ByteBuffer buffer) {
		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();
		final int destinationId = buffer.getInt();

		try {
			putInRingBuffer(new SentRemoteCallRecord(timestamp,
					stringRegistry.getStringFromId(destinationId), traceId, orderIndex,
					hostApplicationMetadata));
		} catch (final IdNotAvailableException e) {
			putInWaitingMessages(buffer, SentRemoteCallRecord.COMPRESSED_BYTE_LENGTH_WITH_CLAZZ_ID);
		}
	}

	private final void readInUnknownReceivedRemoteCallEvent(final ByteBuffer buffer) {
		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();

		final int senderId = buffer.getInt();
		final int destionationId = buffer.getInt();

		try {
			putInRingBuffer(new UnknownReceivedRemoteCallRecord(timestamp,
					stringRegistry.getStringFromId(senderId),
					stringRegistry.getStringFromId(destionationId), traceId, orderIndex,
					hostApplicationMetadata));
		} catch (final IdNotAvailableException e) {
			putInWaitingMessages(buffer,
					UnknownReceivedRemoteCallRecord.COMPRESSED_BYTE_LENGTH_WITH_CLAZZ_ID);
		}
	}

	private final void putInWaitingMessages(final ByteBuffer buffer, final int length) {
		final byte[] message = new byte[length];
		buffer.position(buffer.position() - length);
		buffer.get(message);
		waitingForStringMessages.add(message);
	}

	private final void checkWaitingMessages() {
		final List<byte[]> localWaitingList = new ArrayList<byte[]>();
		for (final byte[] waitingMessage : waitingForStringMessages) {
			localWaitingList.add(waitingMessage);
		}
		waitingForStringMessages.clear();

		for (final byte[] waitingMessage : localWaitingList) {
			final ByteBuffer buffer = ByteBuffer.wrap(waitingMessage);
			final byte waitingMessageClazzId = buffer.get();
			switch (waitingMessageClazzId) {
				case HostApplicationMetaDataRecord.CLAZZ_ID:
					readInHostApplicationMetaData(buffer);
					break;
				case BeforeOperationEventRecord.CLAZZ_ID:
					readInBeforeOperationEvent(buffer);
					break;
				case AfterFailedOperationEventRecord.CLAZZ_ID:
					readInAfterFailedOperationEvent(buffer);
					break;
				case AfterOperationEventRecord.CLAZZ_ID:
					readInAfterOperationEvent(buffer);
					break;
				case BeforeConstructorEventRecord.CLAZZ_ID:
					readInBeforeConstructorEvent(buffer);
					break;
				case AfterFailedConstructorEventRecord.CLAZZ_ID:
					readInAfterFailedConstructorEvent(buffer);
					break;
				case AfterConstructorEventRecord.CLAZZ_ID:
					readInAfterConstructorEvent(buffer);
					break;
				case ReceivedRemoteCallRecord.CLAZZ_ID:
					readInReceivedRemoteCallEvent(buffer);
					break;
				case BeforeStaticOperationEventRecord.CLAZZ_ID:
					readInBeforeStaticOperationEvent(buffer);
					break;
				case AfterFailedStaticOperationEventRecord.CLAZZ_ID:
					readInAfterFailedStaticOperationEvent(buffer);
					break;
				case AfterStaticOperationEventRecord.CLAZZ_ID:
					readInAfterStaticOperationEvent(buffer);
					break;
				case SentRemoteCallRecord.CLAZZ_ID:
					readInSentRemoteCallEvent(buffer);
					break;
				case UnknownReceivedRemoteCallRecord.CLAZZ_ID:
					readInUnknownReceivedRemoteCallEvent(buffer);
					break;
				default:
					break;
			}
		}
	}

	private final void putInRingBuffer(final IRecord message) {
		synchronized (this) { // TODO better solution
			outputBuffer[outputBufferIndex++] = message;
			if (outputBufferIndex == Constants.TCP_READER_OUTPUT_BUFFER_SIZE) {
				flushOutputBuffer();
			}
		}
	}

	public void flushOutputBuffer() {
		synchronized (this) {
			if (outputBufferIndex > 0) {
				final long hiseq = ringBuffer.next();
				final RecordArrayEvent valueEvent = ringBuffer.get(hiseq);
				final IRecord[] oldValues = valueEvent.getValues();
				valueEvent.setValues(outputBuffer);
				valueEvent.setValueSize(outputBufferIndex);
				ringBuffer.publish(hiseq);

				outputBuffer = oldValues;

				outputBufferIndex = 0;
			}
		}
	}

}
